/*
   cvsd - chroot wrapper to run `cvs pserver' more securely.

   cvsd was originally written by Chris Black <cblack@mokey.com>,
   http://cblack.mokey.com/. That was until release 0.6

   cvsd versions up till 0.8b3 were maintained by Philippe Kehl
   <phkehl@gmx.net>, http://guv.ethz.ch/~flip/cvsd/,
   http://www.oinkzwurgl.org/software/cvsd/.

   after that Arthur de Jong <arthur@tiefighter.et.tudelft.nl>
   took up the work and did a complete rewrite,
   http://tiefighter.et.tudelft.nl/~arthur/cvsd.

   Copyright (C) 1999 Chris Black.
   Copyright (C) 2000 Philippe Kehl.
   Copyright (C) 2003 Arthur de Jong.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include "config.h"

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <stdlib.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif /* HAVE_GETOPT_H */
#ifndef HAVE_GETOPT_LONG
#include "getopt_long.h"
#endif /* not HAVE_GETOPT_LONG */
#ifndef HAVE_DAEMON
#include "daemon.h"
#endif /* not HAVE_DAEMON */
#ifndef HAVE_GETADDRINFO
#include "getinfos.h"
#endif /* not HAVE_GETADDRINFO */

#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif /* HAVE_NETDB_H */
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif /* HAVE_FCNTL_H */
#ifdef USE_LIBWRAP
#include <tcpd.h>
#endif /* USE_LIBWRAP */
#ifdef HAVE_GRP_H
#include <grp.h>
#endif /* HAVE_GRP_H */

#include "xmalloc.h"
#ifdef HAVE_SETRLIMIT
#include "reslimit.h"
#endif /* HAVE_SETRLIMIT */
#include "cfgfile.h"
#include "log.h"


/* the definition of the environment */
extern char **environ;


/* the name by which this program was run. */
static char *program_name;


/* the command that will be run */
char *cvs_cmd=NULL;


/* the arguments that will be passed */
static char **cvs_args;


/* the environment cvs will run in */
#define CVSUMASK_IDX 2
static char *cvs_env[]={
  "HOME=/",
  "PATH=/bin",
  "SHELL=/bin/sh",
  "TMPDIR=/tmp",
  NULL, /* this is set to CVSUMASK=<cvs_umask> */
  NULL
};


/* the umask value cvs will run as */
mode_t cvs_umask=0027;


/* the configuration file default from Makefile */
char *cvsd_configfile=DEFAULT_CONFIGFILE;


/* the nice value the daemon should run as */
int cvsd_nice=0;


/* the user id cvs should be run as */
#define NOUID ((gid_t)-1)
uid_t cvsd_uid=NOUID;


/* the group id cvs should be run as */
#define NOGID ((gid_t)-1)
gid_t cvsd_gid=NOGID;


/* cvsd root directory */
char *cvsd_rootjail=NULL;


/* the pid file */
char *cvsd_pidfile=NULL;


/* the addresses to listen on */
static struct cvsd_addrs
{
  char *node;
  char *service;
  struct addrinfo *addrs;
  struct cvsd_addrs *next;
} *cvsd_addresses=NULL;


/* the default address and port to listen on */
#define DEFAULT_ADDR "*"
#define DEFAULT_PORT "2401"


/* the maximum number of connections to handle */
int cvsd_maxconnections=0; /* default=0=unlimited */


/* whether debugging mode is enabled */
static int cvsd_debugging=0;


/* the number of children spawned */
static volatile int cvsd_numchildren=0;


/* the server sockets */
#define MAXSERVERSOCKETS 16
static int cvsd_serversockets[MAXSERVERSOCKETS];
static int cvsd_serversocketnum=0;


/* the number of seconds to sleep when no more
   connections can be listened for */
#define SLEEPSECS 5


/* display version information */
static void display_version(FILE *fp)
{
  fprintf(fp,"%s\n",PACKAGE_STRING);
  fprintf(fp,"Written by Chris Black, Philippe Kehl and Arthur de Jong.\n\n");
  fprintf(fp,"Copyright (C) 1999,2000,2003 Chris Black, Philippe Kehl and Arthur de Jong.\n"
             "This is free software; see the source for copying conditions.  There is NO\n"
             "warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n");
}


/* display usage information to stdout and exit(status) */
static void display_usage(FILE *fp)
{
  fprintf(fp,"Usage: %s [OPTION]...\n",program_name);
  fprintf(fp,"chroot wrapper to run `cvs pserver' more securely.\n");
  fprintf(fp,"  -f, --config=FILE  use FILE as configfile (default %s)\n",DEFAULT_CONFIGFILE);
  fprintf(fp,"  -d, --debug        don't fork and print debugging to stderr\n");
  fprintf(fp,"      --help         display this help and exit\n");
  fprintf(fp,"      --version      output version information and exit\n");
  fprintf(fp,"\n"
             "Report bugs to <%s>.\n",PACKAGE_BUGREPORT);
}


/* the definition of options for getopt(). see getopt(2) */
static struct option const cvsd_options[] =
{
  { "config",      required_argument, NULL, 'f' },
  { "debug",       no_argument,       NULL, 'd' },
  { "help",        no_argument,       NULL, 'h' },
  { "version",     no_argument,       NULL, 'V' },
  { NULL, 0, NULL, 0 }
};
#define CVSD_OPTIONSTRING "f:dhV"


/* parse command line options and set global vars */
static void parse_cmdline(int argc,char *argv[])
{
  int optc;
  while ((optc=getopt_long(argc,argv,CVSD_OPTIONSTRING,cvsd_options,NULL))!=-1)
  {
    switch (optc)
    {
    case 'f': /* -f, --config=FILE  use FILE as configfile */
      if (optarg[0]!='/')
      {
        fprintf(stderr,"%s: configfile '%s' must be an absolute path\n",
                program_name,optarg);
        exit(1);
      }
      cvsd_configfile=xstrdup(optarg);
      break;
    case 'd': /* -d, --debug        don't fork and print debugging to stderr */
      cvsd_debugging=1;
      setdefaultloglevel(LOG_DEBUG);
      break;
    case 'h': /*     --help         display this help and exit */
      display_usage(stdout);
      exit(0);
    case 'V': /*     --version      output version information and exit */
      display_version(stdout);
      exit(0);
    case ':': /* missing required parameter */
    case '?': /* unknown option character or extraneous parameter */
    default:
      fprintf(stderr,"Try `%s --help' for more information.\n",
              program_name);
      exit(1);
    }
  }
  /* check for remaining arguments */
  if (optind<argc)
  {
    fprintf(stderr,"%s: unrecognized option `%s'\n",
            program_name,argv[optind]);
    fprintf(stderr,"Try `%s --help' for more information.\n",
            program_name);
    exit(1);
  }
}


/* add a string as a command line */
void add_cvsarg(char *arg)
{
  static int sz=0; /* size of array */
  static int cn=0; /* number of elements */
  char **tmp;
  int nsz;
  /* first ensure array is large enough */
  if ((cn+2)>sz)
  {
    if (sz==0)
      nsz=10;
    else
      nsz=sz*2;
    tmp=(char **)xmalloc(nsz*sizeof(char *));
    memcpy(tmp,cvs_args,sz*sizeof(char *));
    cvs_args=tmp;
    sz=nsz;
  }
  /* add string to end of array */
  cvs_args[cn++]=arg;
  cvs_args[cn]=NULL;
}


/* get a name of a signal with a given signal number */
static const char *signame(int signum)
{
  switch (signum)
  {
  case SIGHUP:  return "SIGHUP";  /* Hangup detected */
  case SIGINT:  return "SIGINT";  /* Interrupt from keyboard */
  case SIGQUIT: return "SIGQUIT"; /* Quit from keyboard */
  case SIGILL:  return "SIGILL";  /* Illegal Instruction */
  case SIGABRT: return "SIGABRT"; /* Abort signal from abort(3) */
  case SIGFPE:  return "SIGFPE";  /* Floating point exception */
  case SIGKILL: return "SIGKILL"; /* Kill signal */
  case SIGSEGV: return "SIGSEGV"; /* Invalid memory reference */
  case SIGPIPE: return "SIGPIPE"; /* Broken pipe */
  case SIGALRM: return "SIGALRM"; /* Timer signal from alarm(2) */
  case SIGTERM: return "SIGTERM"; /* Termination signal */
  case SIGUSR1: return "SIGUSR1"; /* User-defined signal 1 */
  case SIGUSR2: return "SIGUSR2"; /* User-defined signal 2 */
  case SIGCHLD: return "SIGCHLD"; /* Child stopped or terminated */
  case SIGCONT: return "SIGCONT"; /* Continue if stopped */
  case SIGSTOP: return "SIGSTOP"; /* Stop process */
  case SIGTSTP: return "SIGTSTP"; /* Stop typed at tty */
  case SIGTTIN: return "SIGTTIN"; /* tty input for background process */
  case SIGTTOU: return "SIGTTOU"; /* tty output for background process */
#ifdef SIGBUS
  case SIGBUS:  return "SIGBUS";  /* Bus error */
#endif
#ifdef SIGPOLL
  case SIGPOLL: return "SIGPOLL"; /* Pollable event */
#endif
#ifdef SIGPROF
  case SIGPROF: return "SIGPROF"; /* Profiling timer expired */
#endif
#ifdef SIGSYS
  case SIGSYS:  return "SIGSYS";  /* Bad argument to routine */
#endif
#ifdef SIGTRAP
  case SIGTRAP: return "SIGTRAP"; /* Trace/breakpoint trap */
#endif
#ifdef SIGURG
  case SIGURG:  return "SIGURG";  /* Urgent condition on socket */
#endif
#ifdef SIGVTALRM
  case SIGVTALRM: return "SIGVTALRM"; /* Virtual alarm clock */
#endif
#ifdef SIGXCPU
  case SIGXCPU: return "SIGXCPU"; /* CPU time limit exceeded */
#endif
#ifdef SIGXFSZ
  case SIGXFSZ: return "SIGXFSZ"; /* File size limit exceeded */
#endif
  default:      return "UNKNOWN";
  }
}


/* signal handler for dieing children */
static RETSIGTYPE sigchld_handler(int signum)
{
  int status;
  pid_t cpid;
  /* loop until all pending child processes have finished
     (kill all the zombies) */
  while ((cpid=waitpid(-1,&status,WNOHANG))>0)
  {
    /* shouldn't have a race here because we don't have threads */
    cvsd_numchildren--;
    if (WIFSIGNALED(status))
    {
      log(LOG_INFO,"cvs command exited on signal %s (%d) with exit-status %d",
                   signame((int)WTERMSIG(status)),
                   (int)WTERMSIG(status),
                   (int)WEXITSTATUS(status));
    }
    else
    {
      log(LOG_INFO,"cvs command exited%s with exit-status %d",
                   WIFEXITED(status)?"":" abnormally",
                   (int)WEXITSTATUS(status));
    }
  }
}


/* signal handler for closing down */
static RETSIGTYPE sigexit_handler(int signum)
{
  log(LOG_INFO,"caught signal %s (%d), shutting down",
               signame(signum),signum);
  exit(1);
}


/* do cleaning up before terminating */
static void exithandler(void)
{
  int i;
  for (i=0;i<=cvsd_serversocketnum;i++)
    close(cvsd_serversockets[i]);
  log(LOG_INFO,"version %s bailing out",VERSION);
}


/* handle a connection by doing fork() and stuff */
static void handleconnection(int csock)
{
  pid_t cpid;
  int i,m;

  /* we have a possible race here? (probably atomic) */
  cvsd_numchildren++;

  /* fork off a child to handle the connection */
  switch (cpid=fork())
  {
  case 0: /* we are the child */
    /* set limits */
    setreslimits();
    /* connect socket to stdin/stdout/stderr */
    if ( (dup2(csock,0)<0) || /* stdin */
         (dup2(csock,1)<0) || /* stdout */
         (dup2(csock,2)<0) )  /* stderr */
    {
      log(LOG_ERR,"dup failed: %s",strerror(errno));
      _exit(1);
    }
    /* close all other file descriptors */
    m=sysconf(_SC_OPEN_MAX);
    for (i=3;i<m;i++)
      close(i);
    /* execute cvs */
    execve(cvs_cmd,cvs_args,cvs_env);
    /* if we are here there has been an error */
    /* we can't log since we don't have any useful file descriptors */
    close(0);
    close(1);
    close(2);
    _exit(1);
    break;
  case -1: /* we are the parent, but have an error */
    log(LOG_ERR,"fork() failed: %s",strerror(errno));
    exit(1);
    break;
  }

  /* we are the parent and have a cpid */
  log(LOG_DEBUG,"debug: fork() succeeded (child pid=%d)",(int)cpid);

  /* close the socket to prevent confusion */
  close(csock);
}


/* add the given node and service to the list of addresses that
   will be listened on */
void addaddress(const char *filename,int lnr,
                const char *node,const char *service)
{
  struct cvsd_addrs *addr,*tmp;
  struct addrinfo hints;
  int i;

  /* create space to store address */
  addr=(struct cvsd_addrs *)xmalloc(sizeof(struct cvsd_addrs));
  addr->node=xstrdup(node);
  addr->service=xstrdup(service);
  addr->addrs=NULL;
  addr->next=NULL;

  /* some aliases for the unspecified address */
  if ((strcmp(node,"*")==0)||(strcmp(node,"")==0))
    node=NULL;

  /* create the name for the socket */
  memset(&hints,0,sizeof(struct addrinfo));
  hints.ai_flags=AI_PASSIVE;
  hints.ai_family=PF_UNSPEC;
  hints.ai_socktype=SOCK_STREAM;
  switch (i=getaddrinfo(node,service,&hints,&(addr->addrs)))
  {
    case 0: break;
    case EAI_SYSTEM:
      if (filename==NULL)
        log(LOG_ERR,"getaddrinfo() failed: %s",strerror(errno));
      else
        log(LOG_ERR,"%s:%d: getaddrinfo() failed: %s",
                    filename,lnr,strerror(errno));
      exit(1);
    default:
      if (filename==NULL)
        log(LOG_ERR,"getaddrinfo() failed: %s",gai_strerror(i));
      else
        log(LOG_ERR,"%s:%d: getaddrinfo() failed: %s",
                    filename,lnr,gai_strerror(i));
      exit(1);
  }
  if (addr->addrs==NULL)
  {
    log(LOG_ERR,"%s:%d: getaddrinfo() returned no usable addresses",filename,lnr);
    exit(1);
  }

  /* save the address */
  if (cvsd_addresses==NULL)
    cvsd_addresses=addr;
  else
  {
    for (tmp=cvsd_addresses;tmp->next!=NULL;tmp=tmp->next);
    tmp->next=addr;
  }
}


/* create socket, bind and listen on the specified address */
static void listenonaddresses(const struct cvsd_addrs *addrs)
{
  int flag;
  int sock;
  char host[80],serv[40];
  int i;
  struct addrinfo *last=NULL;
  struct addrinfo *addr;
  int count=0;

  for (addr=addrs->addrs;addr!=NULL;addr=addr->ai_next)
  {

    switch (i=getnameinfo(addr->ai_addr,addr->ai_addrlen,
                          host,80,serv,40,
                          NI_NUMERICHOST|NI_NUMERICSERV)<0)
    {
      case 0:
        log(LOG_DEBUG,"debug: binding %s %s family=%d socktype=%d protocol=%d",host,serv,
                      addr->ai_family,addr->ai_socktype,addr->ai_protocol);
        break;
      case EAI_SYSTEM:
        log(LOG_DEBUG,"debug: getnameinfo() failed: %s",strerror(errno));
        break;
      default:
        log(LOG_DEBUG,"debug: getnameinfo() failed: %s",gai_strerror(i));
        break;
    }

    /* ignore unix domain sockets (i.e. named sockets) */
    if (addr->ai_family==PF_UNIX)
    {
      log(LOG_DEBUG,"debug: do not try unix domain socket (family %d)",addr->ai_family);
      continue;
    }

    /* create socket */
    if ((sock=socket(addr->ai_family,addr->ai_socktype,addr->ai_protocol))<0)
    {
      if ((errno==EAFNOSUPPORT)||(errno==EINVAL))
      {
        log(LOG_DEBUG,"debug: socket() failed (ignored): %s",strerror(errno));
        continue;
      }
      log(LOG_ERR,"cannot create socket: %s",strerror(errno));
      exit(1);
    }

    /* bind socket even if port is in close wait state */
    flag=1;
    if (setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,&flag,sizeof(int))!=0)
    {
      log(LOG_ERR,"setsockopt(SO_REUSEADDR) failed: %s",strerror(errno));
      close(sock);
      exit(1);
    }

#ifdef TODO
    /* close the socket fast (SO_LINGER) */
    struct linger {
                    int   l_onoff;    /* linger active */
                    int   l_linger;   /* how many seconds to linger for */
                };
    if (setsockopt(sock,SOL_SOCKET,SO_LINGER,&**,sizeof(**))!=0)
    {
      log(LOG_ERR,"setsockopt(SO_LINGER) failed: %s",strerror(errno));
      close(sock);
      exit(1);
    }
#endif /* TODO */

    /* do not block on accept() */
    if ((i=fcntl(sock,F_GETFL,0))<0)
    {
      log(LOG_ERR,"fctnl(F_GETFL) failed: %s",strerror(errno));
      close(sock);
      exit(1);
    }
    if (fcntl(sock,F_SETFL,i|O_NONBLOCK)<0)
    {
      log(LOG_ERR,"fctnl(F_SETFL,O_NONBLOCK) failed: %s",strerror(errno));
      close(sock);
      exit(1);
    }

    /* bind the socket to the specified port */
    if (bind(sock,addr->ai_addr,addr->ai_addrlen)<0)
    {
      if ((errno==EADDRINUSE)&&(last!=NULL)&&
          (last->ai_family==AF_INET6)&&
          (addr->ai_family==AF_INET))
      {
        close(sock);
        log(LOG_DEBUG,"debug: bind() failed (ignored): %s",strerror(errno));
        continue;
      }
      log(LOG_ERR,"bind() failed: %s",strerror(errno));
      close(sock);
      exit(1);
    }

    /* start listening for connections */
    if (listen(sock,1)<0)
    {
      log(LOG_ERR,"listen() failed: %s",strerror(errno));
      close(sock);
      exit(1);
    }

    /* add the allocated socket to the list of ok sockets */
    if (cvsd_serversocketnum>=MAXSERVERSOCKETS)
    {
      log(LOG_ERR,"maximum number of listening serversockets reached: edit MAXSERVERSOCKETS in cvsd.c and recompile!");
      close(sock);
      exit(1);
    }
    cvsd_serversockets[cvsd_serversocketnum++]=sock;

    /* log stuff */
    switch (i=getnameinfo(addr->ai_addr,addr->ai_addrlen,
                          host,80,serv,40,
                          NI_NUMERICHOST|NI_NUMERICSERV)<0)
    {
      case 0:
        log(LOG_INFO,"listening on %s %s",host,serv);
        break;
      case EAI_SYSTEM:
        log(LOG_ERR,"getnameinfo() failed (ignored): %s",strerror(errno));
        break;
      default:
        log(LOG_ERR,"getnameinfo() failed (ignored): %s",gai_strerror(i));
        break;
    }

    last=addr;
    count++;
  } /* next addr */

  if (count<=0)
  {
    log(LOG_ERR,"failed binding any address for %s %s: %s",
                addrs->node,addrs->service,strerror(errno));
    exit(1);
  }

}


/* accept a connection on the socket and spawn a child cvs */
static void acceptconnection(void)
{
  int csock;
  fd_set fds;
  int i;
  int j;
  int max=0;
  struct sockaddr_storage addr;
  socklen_t alen;
  char hostbuf[80],serv[40];
  char *host;
#ifdef USE_LIBWRAP
  struct request_info rinfo;
#endif /* USE_LIBWRAP */

  /* clear the set */
  FD_ZERO(&fds);
  for (i=0;i<cvsd_serversocketnum;i++)
  {
    FD_SET(cvsd_serversockets[i],&fds);
    if (cvsd_serversockets[i]+1>max)
      max=cvsd_serversockets[i]+1;
  }

  /* wait for something to change */
  if (select(max,&fds,NULL,NULL,NULL)<0)
  {
    if (errno==EINTR)
    {
      log(LOG_DEBUG,"debug: select() failed (ignored): %s",strerror(errno));
      return;
    }
    log(LOG_ERR,"select() failed: %s",strerror(errno));
    exit(1);
  }

  for (i=0;i<cvsd_serversocketnum;i++)
  {
    if (FD_ISSET(cvsd_serversockets[i],&fds))
    {

      /* accept a new connection */
      alen=(socklen_t)sizeof(struct sockaddr_storage);
      csock=accept(cvsd_serversockets[i],(struct sockaddr *)&addr,&alen);
      if (csock<0)
      {
        if ((errno==EINTR)||(errno==EAGAIN)||(errno==EWOULDBLOCK))
        {
          log(LOG_DEBUG,"debug: accept() failed (ignored): %s",strerror(errno));
          continue; /* this is 'normal' */
        }
        log(LOG_ERR,"accept() failed: %s",strerror(errno));
        continue;
      }

      /* look up other side of connection and log it */
      host=hostbuf;
      switch (j=getnameinfo((struct sockaddr *)&addr,alen,
                            host,80,serv,40,
                            NI_NUMERICHOST|NI_NUMERICSERV)<0)
      {
        case 0: break;
        case EAI_SYSTEM:
          log(LOG_ERR,"getnameinfo() failed: %s",strerror(errno));
          close(csock);
          continue;
        default:
          log(LOG_ERR,"getnameinfo() failed: %s",gai_strerror(j));
          close(csock);
          continue;
      }
      /* convert IPv4-mapped-IPv6 address to an IPv4 one */
      if ((strncmp(host,"::ffff:",7)==0)&&
          (strchr(host+7,':')==NULL))
        host+=7;

#ifdef USE_LIBWRAP
      /* find out what hosts.{allow,deny} think about this */
      request_init(&rinfo,RQ_FILE,csock,RQ_DAEMON,PACKAGE,0);
      fromhost(&rinfo); /* not documented, but needed somehow */
      if(!hosts_access(&rinfo))
      {
        log(LOG_INFO,"connection from %s %s refused by tcp wrappers",host,serv);
        close(csock);
        continue;
      }
#endif /* USE_LIBWRAP */

      /* log connection */
      log(LOG_INFO,"connection from %s %s",host,serv);

      /* make sure O_NONBLOCK is not inherited */
      if ((j=fcntl(csock,F_GETFL,0))<0)
      {
        log(LOG_ERR,"fctnl(F_GETFL) failed: %s",strerror(errno));
        close(csock);
        exit(1);
      }
      if (fcntl(csock,F_SETFL,j&~O_NONBLOCK)<0)
      {
        log(LOG_ERR,"fctnl(F_SETFL,~O_NONBLOCK) failed: %s",strerror(errno));
        close(csock);
        exit(1);
      }

      /* handle the connection */
      handleconnection(csock);

    }
  }
}


/* write the current process id to the specified file */
static void write_pidfile(const char *filename)
{
  FILE *fp;
  if (filename!=NULL)
  {
    umask(0022);
    if ((fp=fopen(filename,"w"))==NULL)
    {
      log(LOG_ERR,"cannot open pid file (%s): %s",filename,strerror(errno));
      exit(1);
    }
    fprintf(fp,"%d\n",(int)getpid());
    fclose(fp);
  }
}


/* try to install signal handler and check result */
static void install_sighandler(int signum,RETSIGTYPE (*handler) (int))
{
  struct sigaction act;
  memset(&act,0,sizeof(struct sigaction));
  act.sa_handler=handler;
  sigemptyset(&act.sa_mask);
  act.sa_flags=SA_RESTART|SA_NOCLDSTOP;
  if (sigaction(signum,&act,NULL)!=0)
  {
    log(LOG_ERR,"error installing signal handler for '%s': %s",signame(signum),strerror(errno));
    exit(1);
  }
}


/* the main program... */
int main(int argc,char *argv[])
{
  int i;
  int j;
  char *cvs_cmdline=NULL;
  struct cvsd_addrs *tmp,*nxt;

  /* save program name */
  program_name=argv[0];

  /* clear the environment */
  environ=cvs_env;

  /* parse the command line */
  parse_cmdline(argc,argv);

  /* begin building command line */
  add_cvsarg("cvs");
  add_cvsarg("-f"); /* do not read ~/.cvsrc */
#ifdef DEBUG
  add_cvsarg("-t"); /* verbose */
#endif /* DEBUG */

  /* read the configuration file */
  read_configfile(cvsd_configfile);

  /* specify cvs_command if not set earlier */
  if (cvs_cmd==NULL)
  {
    cvs_cmd=( (cvsd_rootjail!=NULL) &&
              (strcmp(cvsd_rootjail,"none")!=0) )?"/bin/cvs":CVS_LOCATION;
  }
  log(LOG_DEBUG,"debug: cvs_cmd: %s",cvs_cmd);

  /* add cvs action */
  add_cvsarg("pserver"); /* run cvs as pserver */
  for (i=0;cvs_args[i]!=NULL;i++)
    log(LOG_DEBUG,"debug: cvs_args[%d]: %s",i,cvs_args[i]);

  /* complete environment for cvs */
  cvs_env[CVSUMASK_IDX]=(char *)xmalloc(13*sizeof(char));
  i=snprintf(cvs_env[CVSUMASK_IDX],13,"CVSUMASK=%03o",cvs_umask);
  if ( (i<0) || (i>12) )
  {
    log(LOG_ERR,"fatal error saving umask in environment (probably a bug)");
    exit(1);
  }
  log(LOG_DEBUG,"debug: cvs_env[CVSUMASK_IDX]: %s",cvs_env[CVSUMASK_IDX]);

  /* daemonize */
  if ((!cvsd_debugging)&&(daemon(0,0)<0))
  {
    log(LOG_ERR,"unable to daemonize: %s",strerror(errno));
    exit(1);
  }

  /* intilialize logging */
  if (!cvsd_debugging)
    startlogging();
#ifdef DEBUG
  log(LOG_INFO,"version %s starting (debugging enabled)",VERSION);
#else /* DEBUG */
  log(LOG_INFO,"version %s starting",VERSION);
#endif /* not DEBUG */

  /* install handler to close stuff off on exit and log notice */
  atexit(exithandler);

  /* write pidfile */
  write_pidfile(cvsd_pidfile);

  /* set the umask */
  umask(cvs_umask);

  /* default address */
  if (cvsd_addresses==NULL)
    addaddress(NULL,0,DEFAULT_ADDR,DEFAULT_PORT);

  /* start listening on addresses from configfile */
  for (tmp=cvsd_addresses;tmp!=NULL;tmp=nxt)
  {
    listenonaddresses(tmp);
    nxt=tmp->next;
    freeaddrinfo(tmp->addrs);
    free(tmp->node);
    free(tmp->service);
    free(tmp);
  }
  cvsd_addresses=NULL;

  /* do the chroot stuff */
  if ( (cvsd_rootjail!=NULL) &&
       (strcmp(cvsd_rootjail,"none")!=0) )
  {
    if (chdir(cvsd_rootjail)!=0)
    {
      log(LOG_ERR,"cannot chdir(%s): %s",cvsd_rootjail,strerror(errno));
      exit(1);
    }
    if (chroot(cvsd_rootjail)!=0)
    {
      log(LOG_ERR,"cannot chroot(%s): %s",cvsd_rootjail,strerror(errno));
      exit(1);
    }
    log(LOG_DEBUG,"debug: chroot(%s) done",cvsd_rootjail);
  }
  /* just to be sure */
  chdir("/");

  /* renice */
  errno=0; /* for strange nice implementations values */
  nice(cvsd_nice);
  if (errno!=0)
  {
    log(LOG_ERR,"cannot nice(%d): %s",(int)cvsd_nice,strerror(errno));
    exit(1);
  }
  log(LOG_DEBUG,"debug: nice(%d) done",(int)cvsd_nice);

#ifdef HAVE_SETGROUPS
  /* drop all supplemental groups */
  if (setgroups(0,NULL)<0)
  {
    log(LOG_WARNING,"cannot setgroups(0,NULL) (ignored): %s",strerror(errno));
  }
#endif /* HAVE_SETGROUPS */

  /* change to cvs gid */
  if (cvsd_gid!=NOGID)
  {
    if (setgid(cvsd_gid)!=0)
    {
      log(LOG_ERR,"cannot setgid(%d): %s",(int)cvsd_gid,strerror(errno));
      exit(1);
    }
    log(LOG_DEBUG,"debug: setgid(%d) done",cvsd_gid);
  }

  /* change to cvs uid */
  if (cvsd_uid!=NOUID)
  {
    if (setuid(cvsd_uid)!=0)
    {
      log(LOG_ERR,"cannot setuid(%d): %s",(int)cvsd_uid,strerror(errno));
      exit(1);
    }
    log(LOG_DEBUG,"debug: setuid(%d) done",cvsd_uid);
  }

  /* build cvs command line */
  j=strlen(cvs_cmd)+1;
  for (i=1;cvs_args[i]!=NULL;i++)
    j+=strlen(cvs_args[i])+1;
  cvs_cmdline=(char *)xmalloc(j*sizeof(char));
  strcpy(cvs_cmdline,cvs_cmd);
  for (i=1;cvs_args[i]!=NULL;i++)
  {
    strcat(cvs_cmdline," ");
    strcat(cvs_cmdline,cvs_args[i]);
  }
  log(LOG_DEBUG,"debug: cvs command to execute: '%s'",cvs_cmdline);

  /* install signalhandler for terminating children */
  install_sighandler(SIGCHLD,sigchld_handler);

  /* install signalhandlers for some other signals */
  install_sighandler(SIGHUP, sigexit_handler);
  install_sighandler(SIGINT, sigexit_handler);
  install_sighandler(SIGQUIT,sigexit_handler);
  install_sighandler(SIGILL, sigexit_handler);
  install_sighandler(SIGABRT,sigexit_handler);
  install_sighandler(SIGSEGV,sigexit_handler);
  install_sighandler(SIGPIPE,sigexit_handler);
  install_sighandler(SIGALRM,sigexit_handler);
  install_sighandler(SIGTERM,sigexit_handler);
  install_sighandler(SIGUSR1,sigexit_handler);
  install_sighandler(SIGUSR2,sigexit_handler);

  log(LOG_INFO,"accepting connections");

  /* start waiting for incoming connections */
  while (1)
  {
    /* if we are ready to handle a connection do so */
    if ((cvsd_maxconnections>0)&&(cvsd_numchildren<cvsd_maxconnections))
    {
      /* wait for a new connection */
      acceptconnection();
    }
    else /* not ready for connection */
    {
      /* log that the connections are full */
      log(LOG_INFO,"too many connections (%d), not accepting new connections",cvsd_numchildren);
      /* wait until children have died */
      while (cvsd_numchildren>=cvsd_maxconnections)
      {
        /* sleep unblocks on a signal */
        sleep(SLEEPSECS);
      }
      /* log that the connections are full */
      log(LOG_INFO,"resuming accepting new connections");
    }
  }

  return 1; /* keep compiler happy */
}
